"""
Implements a curator-catalogue to CMS' DAS based on the dasgoclient command
"""
import subprocess
from .common import check_entries_uproot


_dasgoclient_prog = "dasgoclient"
_proxyinfo_prog = "voms-proxy-info"
_default_prefix = "root://cms-xrd-global.cern.ch//"


def _check_proxy():
    try:
        check_proxy = subprocess.run([_proxyinfo_prog], capture_output=True, text=True)
    except FileNotFoundError as e:
        if _proxyinfo_prog in str(e):
            msg = "%s not found. Needed to test voms proxy"
            return RuntimeError(msg % _proxyinfo_prog)
    if "Proxy not found" in check_proxy.stderr:
        msg = "No valid VOMS proxy configured.  Please run `voms-proxy-init --voms cms`"
        return RuntimeError(msg)


def _check_help():
    try:
        subprocess.run([_dasgoclient_prog, "-h"], capture_output=True)
    except FileNotFoundError as e:
        if _dasgoclient_prog in str(e):
            msg = "%s program not found, please set up necessary CMS environment"
            return RuntimeError(msg % _dasgoclient_prog)


def check_setup():
    error = _check_help()
    if error:
        raise error

    error = _check_proxy()
    if error:
        raise error
    return True


def expand_file_list(datasets, prefix=None):
    if not prefix:
        prefix = _default_prefix

    files = []
    for dataset in datasets:
        query = "-query=file dataset={}".format(dataset)
        cmd = [_dasgoclient_prog, query, "-limit", "0", "-unique"]
        das_result = subprocess.run(cmd, capture_output=True, text=True)
        files += [prefix + f for f in das_result.stdout.split("\n") if f]
    return files


def check_files(*args, **kwargs):
    return check_entries_uproot(*args, **kwargs)



"""Top-level package for fast-curator."""
from . import read
from . import write
from .version import __version__, version_info


__all__ = ["read", "write", "__version__", "version_info"]

__author__ = """Benjamin Krikler, and F.A.S.T"""
__email__ = 'fast-hep@cern.ch'



import os
import sys
import uproot
from collections import defaultdict, Counter
from functools import partial
if sys.version[0] > '2':
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


class XrootdExpander():
    """
    Expand wild-carded file paths, including with xrootd-served files
    """
    @staticmethod
    def check_setup():
        return True

    @staticmethod
    def expand_file_list(files, prefix=None):
        from XRootD.client.glob_funcs import glob
        return expand_file_list_generic(files, prefix,
                                        glob=partial(glob, raise_error=True))

    @staticmethod
    def check_files(*args, **kwargs):
        return check_entries_uproot(*args, **kwargs)


class LocalGlobExpander():
    """
    Expand wild-carded file paths on the local file system
    """
    import glob

    @staticmethod
    def check_setup():
        return True

    @staticmethod
    def expand_file_list(files, prefix=None):
        glob = LocalGlobExpander.glob.glob
        return expand_file_list_generic(files, prefix, glob=glob)

    @staticmethod
    def check_files(*args, **kwargs):
        return check_entries_uproot(*args, **kwargs)


def expand_file_list_generic(files, prefix, glob):
    full_list = []
    for name in files:
        scheme = urlparse(name).scheme
        if not scheme and not os.path.isabs(name):
            if prefix:
                name = os.path.join(prefix, name)
            else:
                name = os.path.relpath(name)
        expanded = glob(name)
        full_list += map(str, expanded)
    return full_list


def check_entries_uproot(files, tree_names, no_empty, confirm_tree=True, list_branches=False,
                         ignore_inaccessible=False):
    no_empty = no_empty or confirm_tree
    if not isinstance(tree_names, (tuple, list)):
        tree_names = [tree_names]

    if ignore_inaccessible:
        files = [f for f in files if os.access(f, os.R_OK)]

    if not no_empty:
        n_entries = {tree: uproot.numentries(files, tree) for tree in tree_names}
    else:
        n_entries = {tree: 0 for tree in tree_names}
        missing_trees = defaultdict(list)
        for tree in tree_names:
            totals = uproot.numentries(files, tree, total=False)
            for name, entries in totals.items():
                n_entries[tree] += entries
                if no_empty and entries <= 0:
                    files.remove(name)
                if confirm_tree and entries == 0:
                    if tree not in uproot.open(name):
                        missing_trees[tree].append(name)
        if missing_trees:
            files = set(sum((list(v) for v in missing_trees.values()), []))
            msg = "Missing at least one tree (%s) for %d file(s): %s"
            msg = msg % (", ".join(missing_trees), len(missing_trees), ", ".join(files))
            raise RuntimeError(msg)

    branches = {}
    if list_branches:
        for tree in tree_names:
            open_files = (uproot.open(f) for f in files)
            all_branches = (f[tree].keys(recursive=True) for f in open_files if tree in f)
            branches[tree] = dict(Counter(sum(all_branches, [])))

    if len(n_entries) == 1:
        n_entries = list(n_entries.values())[0]
    return files, n_entries, branches



import six
import sys
import os

if sys.version_info[0] >= 3.3:
    from types import SimpleNamespace as Dataset
else:
    from argparse import Namespace as Dataset


def associate_by_ext_suffix(datasets):
    has_ext = []
    has_no_ext = []
    for dataset in datasets:
        out_list = has_ext if "_ext" in dataset.name else has_no_ext
        out_list.append(dataset)

    for dataset in has_no_ext:
        associates = [d for d in has_ext if d.name.startswith(dataset.name)]
        associates.append(dataset)
        names = [a.name for a in associates]
        for index in range(len(associates)):
            associates[index].associates = names[:index]
            associates[index].associates += names[index + 1:]


def _load_yaml(path):
    import yaml
    with open(path, 'r') as f:
        datasets_dict = yaml.safe_load(f)
    if not datasets_dict:
        raise RuntimeError("Empty config file in '%s'" % path)
    return datasets_dict


def from_yaml(path, defaults={}, find_associates=associate_by_ext_suffix,
              selected_prefix=None, expand_prefix=True):
    datasets_dict = _load_yaml(path)
    this_dir = os.path.dirname(os.path.abspath(path))
    return get_datasets(datasets_dict, defaults, this_dir=this_dir,
                        selected_prefix=selected_prefix,
                        expand_prefix=expand_prefix,
                        find_associates=associate_by_ext_suffix)


def get_datasets(datasets_dict, defaults={},
                 find_associates=associate_by_ext_suffix, already_imported=None,
                 this_dir=None, selected_prefix=None, expand_prefix=True):
    datasets = []
    defaults.update(datasets_dict.get("defaults", {}))
    if "import" not in datasets_dict and "datasets" not in datasets_dict:
        raise RuntimeError("Neither 'datasets' nor 'import' were specified in config")

    if already_imported is None:
        already_imported = set()
    for import_file in datasets_dict.get("import", []):
        if this_dir:
            import_file = import_file.format(this_dir=this_dir)
        if import_file in already_imported:
            continue
        already_imported.add(import_file)
        contents = _load_yaml(import_file)
        datasets += get_datasets(contents, defaults=defaults.copy(), this_dir=os.path.dirname(import_file),
                                 find_associates=find_associates, already_imported=already_imported)
    for dataset in datasets_dict.get("datasets", []):
        if isinstance(dataset, six.string_types):
            cfg = _from_string(dataset, defaults)
        elif isinstance(dataset, dict):
            cfg = _from_dict(dataset, defaults, selected_prefix)
        else:
            raise TypeError("{} not a string or dict".format(dataset))
        if expand_prefix:
            prefix = cfg.get("prefix", None)
            files = apply_prefix(prefix, cfg["files"], selected_prefix, cfg["name"])
            cfg["files"] = files
        datasets.append(Dataset(**cfg))

    # Associate samples
    find_associates(datasets)

    return datasets


def _from_string(dataset, default):
    cfg = default.copy()
    cfg["name"] = dataset
    return cfg


def _from_dict(dataset, default, selected_prefix=None):
    cfg = default.copy()
    cfg.update(dataset)
    if "name" not in cfg:
        raise RuntimeError(
            "Dataset provided as dict, without key-value pair for 'name'")
    return cfg


def apply_prefix(prefix, files, selected_prefix, dataset):
    if not prefix:
        return files

    if isinstance(prefix, list):
        if not all((isinstance(p, dict) and len(p) == 1 for p in prefix)):
            raise ValueError("'prefix' is a list, but not all elements are single-length dicts")
        prefix = [tuple(p.items())[0] for p in prefix]
        if selected_prefix:
            matched = [v for p, v in prefix if p == selected_prefix]
            if len(matched) > 1:
                msg = "Prefix '%s' is defined %d times, not sure which to use"
                raise ValueError(msg % (selected_prefix, len(matched)))
            if not matched:
                msg = "Prefix '%s' is not defined for dataset '%s'"
                raise ValueError(msg % (selected_prefix, dataset))
            prefix = matched[0]
        else:
            prefix = prefix[0][1]
    elif not isinstance(prefix, six.string_types):
        msg = "'prefix' for dataset '%s' is type '%s'. Need a string or a list of single-length dicts"
        raise ValueError(msg % (dataset, type(prefix)))

    return [f.format(prefix=prefix) for f in files]



from __future__ import print_function
import six
import itertools
import operator
import os
from . import read
from .catalogues import get_file_list_expander, known_expanders
from collections import defaultdict
import logging
logger = logging.getLogger(__name__)


__all__ = ["known_expanders", "prepare_file_list", "write_yaml",
           "add_meta", "process_user_function"]


def prepare_file_list(files, dataset, eventtype, tree_name, expand_files="xrootd",
                      prefix=None, no_empty_files=True, confirm_tree=True,
                      ignore_inaccessible=False,
                      include_branches=False):
    """
    Expands all globs in the file lists and creates a dataframe similar to those from a DAS query
    """
    if isinstance(expand_files, six.string_types):
        expand_files = get_file_list_expander(expand_files)
    if isinstance(files, six.string_types):
        files = [files]
    full_list = expand_files.expand_file_list(files, prefix=prefix)
    full_list = [os.path.realpath(f) if ':' not in f else f for f in full_list]
    full_list, numentries, branches = expand_files.check_files(full_list, tree_name,
                                                               no_empty=no_empty_files,
                                                               list_branches=include_branches,
                                                               confirm_tree=confirm_tree,
                                                               ignore_inaccessible=ignore_inaccessible)

    data = {}
    if prefix:
        full_list = ["{prefix}" + path[len(prefix):] if path.startswith(prefix) else path for path in full_list]
        data["prefix"] = [{"default": prefix}]
    data["eventtype"] = eventtype
    data["name"] = dataset
    data["nevents"] = numentries
    data["nfiles"] = len(full_list)
    data["files"] = full_list
    data["tree"] = tree_name[0] if len(tree_name) == 1 else tree_name
    if branches:
        data["branches"] = branches

    return data


def select_default(values):
    groups = itertools.groupby(sorted(values))
    groups = [(group, sum(1 for _ in items)) for group, items in groups]
    groups = [group for group in groups if group[1] > 1]
    if not groups:
        return None
    most_common, number_items = max(groups, key=operator.itemgetter(1))
    is_unique = sum([1 for group in groups if group[1] == number_items]) == 1
    if not is_unique:
        return None
    return most_common


def prepare_contents(datasets, no_defaults_in_output=False):
    datasets = [vars(data) if isinstance(data, read.Dataset)
                else data for data in datasets]
    for d in datasets:
        if "associates" in d:
            del d["associates"]

    if no_defaults_in_output:
        # do not group common settings together in default block
        return dict(datasets=datasets)

    # build the default properties
    values = defaultdict(list)
    for data in datasets:
        for k, v in data.items():
            values[k].append(v)

    defaults = {}
    for key, vals in values.items():
        if key == "name":
            continue
        is_in_all_datasets = len(vals) == len(datasets)
        if not is_in_all_datasets:
            continue
        new_default = select_default(vals)
        if new_default:
            defaults[key] = new_default

    cleaned_datasets = []
    for data in datasets:
        new_data = {}
        for key, val in data.items():
            if key in defaults and val == defaults[key]:
                continue
            new_data[key] = val
        cleaned_datasets.append(new_data)

    contents = dict(datasets=cleaned_datasets)
    if defaults:
        contents["defaults"] = defaults
    return contents


def write_yaml(dataset, out_file, append=True, no_defaults_in_output=False):
    import yaml
    if os.path.exists(out_file) and append:
        datasets = read.from_yaml(out_file, expand_prefix=False)
        datasets.append(dataset)
        contents = prepare_contents(datasets, no_defaults_in_output=no_defaults_in_output)
    else:
        contents = {}
        contents["datasets"] = [dataset]

    # https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
    class MyDumper(yaml.Dumper):
        def increase_indent(self, flow=False, indentless=False):
            return super(MyDumper, self).increase_indent(flow, False)

        # disable aliases and anchors, see https://github.com/yaml/pyyaml/issues/103
        def ignore_aliases(self, data):
            return True

    yaml_contents = yaml.dump(
        contents, Dumper=MyDumper, default_flow_style=False)
    with open(out_file, 'w') as out:
        out.write(yaml_contents)

    return yaml_contents


def add_meta(dataset, meta):
    for key, value in meta:
        if key in dataset:
            msg = "Meta data '%s' will override an existing value" % key
            raise RuntimeError(msg)
        dataset[key] = value


def process_user_function(dataset, user_func):
    import importlib
    path = user_func.split(".")
    mod_name = ".".join(path[:-1])
    module = importlib.import_module(mod_name)

    func_name = path[-1]
    function = getattr(module, func_name)

    function(dataset)



from .common import XrootdExpander, LocalGlobExpander
from . import cms_das as CMSDASExpander


known_expanders = dict(xrootd=XrootdExpander,
                       local=LocalGlobExpander,
                       cmsdas=CMSDASExpander,
                       )


def get_file_list_expander(expander):
    if expander not in known_expanders:
        msg = "Unknown catalogue interface requested, '%s'. Valid options: %s"
        raise RuntimeError(msg % (expander, ", ".join(known_expanders.keys())))
    result = known_expanders[expander]
    if not result.check_setup():
        msg = "Issue setting up catalogue interface: %s"
        raise RuntimeError(msg % expander)
    return result
